/*
 * Copyright 2015 - 2024 gary@drinkingtea.net
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 */

#pragma once

#include "ignore.hpp"
#include "memops.hpp"
#include "ox/std/error.hpp"
#include "stringview.hpp"
#include "typetraits.hpp"

namespace ox {

// Inline String
template<std::size_t StrCap>
class IString {
	private:
		char m_buff[StrCap + 1];
		size_t m_size{};

	public:
		constexpr IString() noexcept;

		constexpr IString(StringView str) noexcept;

		constexpr IString(const char *str) noexcept;

		constexpr IString &operator=(CRStringView str) noexcept;

		constexpr IString &operator=(const char *str) noexcept;

		constexpr IString &operator=(char *str) noexcept;

		constexpr IString &operator=(Integer_c auto i) noexcept;

		constexpr IString &operator+=(const char *str) noexcept;

		constexpr IString &operator+=(char *str) noexcept;

		constexpr IString &operator+=(Integer_c auto i) noexcept;

		constexpr IString &operator+=(StringView s) noexcept;

		constexpr IString operator+(const char *str) const noexcept;

		constexpr IString operator+(char *str) const noexcept;

		constexpr IString operator+(Integer_c auto i) const noexcept;

		constexpr bool operator==(const char *other) const noexcept;

		constexpr bool operator==(const OxString_c auto &other) const noexcept;

		constexpr bool operator!=(const char *other) const noexcept;

		constexpr bool operator!=(const OxString_c auto &other) noexcept;

		constexpr char operator[](std::size_t i) const noexcept;

		constexpr char &operator[](std::size_t i) noexcept;

		constexpr Error append(const char *str, std::size_t strLen) noexcept;

		[[nodiscard]]
		constexpr const char *data() const noexcept;

		[[nodiscard]]
		constexpr char *data() noexcept;

		[[nodiscard]]
		constexpr const char *c_str() const noexcept;

		/**
		 * Returns the number of characters in this string.
		 */
		[[nodiscard]]
		constexpr std::size_t len() const noexcept;

		/**
		 * Returns the number of bytes used for this string.
		 */
		[[nodiscard]]
		constexpr std::size_t bytes() const noexcept;

		constexpr ox::Error resize(size_t sz) noexcept;

		/**
		 * Returns the capacity of bytes for this string.
		 */
		[[nodiscard]]
		constexpr std::size_t cap() const noexcept;
};

template<std::size_t size>
constexpr IString<size>::IString() noexcept: m_buff{{0}} {
}

template<std::size_t size>
constexpr IString<size>::IString(StringView str) noexcept: m_buff{{0}} {
	operator=(str);
}

template<std::size_t size>
constexpr IString<size>::IString(const char *str) noexcept: m_buff{{0}} {
	operator=(str);
}

template<std::size_t size>
constexpr IString<size> &IString<size>::operator=(Integer_c auto i) noexcept {
	char str[65] = {};
	ox::itoa(i, str);
	return this->operator=(str);
}

template<std::size_t size>
constexpr IString<size> &IString<size>::operator=(ox::CRStringView str) noexcept {
	std::size_t strLen = str.bytes() + 1;
	if (cap() < strLen) {
		strLen = cap();
	}
	m_size = strLen;
	ox::listcpy(m_buff, str.data(), strLen);
	// make sure last element is a null terminator
	m_buff[strLen] = 0;
	return *this;
}

template<std::size_t size>
constexpr IString<size> &IString<size>::operator=(const char *str) noexcept {
	std::size_t strLen = ox::strlen(str) + 1;
	if (cap() < strLen) {
		strLen = cap();
	}
	m_size = strLen;
	ox::listcpy(m_buff, str, strLen);
	// make sure last element is a null terminator
	m_buff[cap()] = 0;
	return *this;
}

template<std::size_t size>
constexpr IString<size> &IString<size>::operator=(char *str) noexcept {
	return *this = static_cast<const char*>(str);
}

template<std::size_t size>
constexpr IString<size> &IString<size>::operator+=(const char *str) noexcept {
	std::size_t strLen = ox::strlen(str) + 1;
	std::ignore = append(str, strLen);
	return *this;
}

template<std::size_t size>
constexpr IString<size> &IString<size>::operator+=(char *str) noexcept {
	return *this += static_cast<const char*>(str);
}

template<std::size_t size>
constexpr IString<size> &IString<size>::operator+=(Integer_c auto i) noexcept {
	char str[65] = {};
	ox::itoa(i, str);
	return this->operator+=(str);
}

template<std::size_t size>
constexpr IString<size> &IString<size>::operator+=(StringView s) noexcept {
	std::size_t strLen = s.bytes();
	std::ignore = append(s.data(), strLen);
	return *this;
}

template<std::size_t size>
constexpr IString<size> IString<size>::operator+(const char *str) const noexcept {
	auto out = *this;
	std::size_t strLen = ox::strlen(str) + 1;
	std::ignore = out.append(str, strLen);
	return out;
}

template<std::size_t size>
constexpr IString<size> IString<size>::operator+(char *str) const noexcept {
	return *this + static_cast<const char*>(str);
}

template<std::size_t size>
constexpr IString<size> IString<size>::operator+(Integer_c auto i) const noexcept {
	char str[65] = {};
	ox::itoa(i, str);
	return this->operator+(str);
}

template<std::size_t StrCap>
constexpr bool IString<StrCap>::operator==(const char *other) const noexcept {
	return ox::StringView(*this) == other;
}

template<std::size_t StrCap>
constexpr bool IString<StrCap>::operator==(const OxString_c auto &other) const noexcept {
	return ox::StringView(*this) == ox::StringView(other);
}

template<std::size_t StrCap>
constexpr bool IString<StrCap>::operator!=(const char *other) const noexcept {
	return !operator==(other);
}

template<std::size_t StrCap>
constexpr bool IString<StrCap>::operator!=(const OxString_c auto &other) noexcept {
	return !operator==(other);
}

template<std::size_t StrCap>
constexpr char IString<StrCap>::operator[](std::size_t i) const noexcept {
	return m_buff[i];
}

template<std::size_t StrCap>
constexpr char &IString<StrCap>::operator[](std::size_t i) noexcept {
	return m_buff[i];
}

template<std::size_t StrCap>
constexpr Error IString<StrCap>::append(const char *str, std::size_t strLen) noexcept {
	Error err;
	auto currentLen = len();
	if (cap() < currentLen + strLen + 1) {
		strLen = cap() - currentLen;
		err = OxError(1, "Insufficient space for full string");
	}
	ox::strncpy(m_buff + currentLen, str, strLen);
	// make sure last element is a null terminator
	m_buff[currentLen + strLen] = 0;
	return err;
}

template<std::size_t StrCap>
constexpr const char *IString<StrCap>::data() const noexcept {
	return static_cast<const char*>(m_buff);
}

template<std::size_t StrCap>
constexpr char *IString<StrCap>::data() noexcept {
	return static_cast<char*>(m_buff);
}

template<std::size_t StrCap>
constexpr const char *IString<StrCap>::c_str() const noexcept {
	return static_cast<const char*>(m_buff);
}


template<std::size_t StrCap>
constexpr std::size_t IString<StrCap>::len() const noexcept {
	std::size_t length = 0;
	for (std::size_t i = 0; i < StrCap; i++) {
		uint8_t b = static_cast<uint8_t>(m_buff[i]);
		if (b) {
			const auto asciiChar = (b & 128) == 0;
			const auto utf8Char = (b & (256 << 6)) == (256 << 6);
			if (asciiChar || utf8Char) {
				length++;
			}
		} else {
			break;
		}
	}
	return length;
}

template<std::size_t StrCap>
constexpr std::size_t IString<StrCap>::bytes() const noexcept {
	std::size_t i = 0;
	for (i = 0; i < StrCap && m_buff[i]; i++);
	return i + 1; // add one for null terminator
}

template<std::size_t StrCap>
constexpr ox::Error IString<StrCap>::resize(size_t sz) noexcept {
	if (sz > StrCap) {
		return OxError(1, "Trying to extend IString beyond its cap");
	}
	m_size = sz;
	return {};
}

template<std::size_t StrCap>
constexpr std::size_t IString<StrCap>::cap() const noexcept {
	return StrCap;
}

template<size_t sz>
struct MaybeView<ox::IString<sz>> {
	using type = ox::StringView;
};

}
